/*
 * Get PTP time
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <time.h>
#include <unistd.h>

#define NSEC_PER_SEC UINT64_C(1000000000)
#define MAX_BUFFER_SIZE 10000000

// Structs

typedef struct thread_param {
  int interval;
  int priority;

} thread_param_t;

// Static functions

static void process_options(int argc, char *argv[]);
static void add_ns(struct timespec *t, uint64_t ns);
static uint64_t ts_to_uint(struct timespec t);

// Static variables

static thread_param_t thread_params;
static uint64_t ptptime;
static int quit;
static int nowait;
static int print_newline;
static int print_period;
static clockid_t clock_id;

static uint64_t ts_buffer[MAX_BUFFER_SIZE];

static void help(char *argv[]) {
  printf(
      "Usage: %s [-h] [-i USEC]\n"
      "       -h                 Show help\n"
      "       -m                 Get time from CLOCK_MONOTONIC\n"
      "       -r                 Get time from CLOCK_REALTIME (default)\n"
      "       -t                 Get time from CLOCK_TAI\n"
      "       -q                 Get time then quit\n"
      "       -L                 Print new line for each value\n"
      "       -R USEC            Print period\n"
      "       -F                 Read clock and store in buffer as fast as possible and dump the buffer at the end\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "\n",
      argv[0]);
}

/*
 * Real-time thread
 */
static void *rt_thread(void *p) {
  (void)p;
  struct timespec next, ptptime_ts;
  cpu_set_t mask;

  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(1, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU #1\n");

  clock_gettime(CLOCK_MONOTONIC, &next);

  // Time lookup loop
  for (int i = 0;; i++) {

    clock_gettime(clock_id, &ptptime_ts);
    ptptime = ts_to_uint(ptptime_ts);

    if(!nowait) {
      add_ns(&next, thread_params.interval);
      clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next, NULL);
    }
    else {
      ts_buffer[i] = ptptime;

      if((i+1) >= MAX_BUFFER_SIZE) {
        for(int j = 0; j < MAX_BUFFER_SIZE; j++)
          printf("%" PRIu64 "\n", ts_buffer[j]);
        exit(EXIT_SUCCESS);
      }

    }
  }

  return NULL;
}

/*
 * Main thread, has non-real time priority
 * Handles the IO and creates the real time thread
 */
int main(int argc, char *argv[]) {
  pthread_t thread;
  struct sched_param param;
  pthread_attr_t attr;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.priority = 98;
  clock_id = CLOCK_REALTIME;
  print_newline = 0;
  print_period = 100000;

  /* Lock all current and future pages from preventing of being paged to
   * swap */
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    perror("mlockall failed");
    /* exit(-1) or do error handling */
  }

  // Process bash options
  process_options(argc, argv);

  if (quit) {
    struct timespec ts;
    clock_gettime(clock_id, &ts);
    printf("%" PRIu64 "\n", ts_to_uint(ts));
    exit(EXIT_SUCCESS);
  }

  /* Initialize pthread attributes (default values) */
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  /* Set a specific stack size  */
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  /* Set scheduler policy and priority of pthread */
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, rt_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create rt thread");

  // Verbose loop
  for (;;) {
    usleep(print_period);
    if(!nowait) {
      printf("%9" PRIu64 "\n", ptptime);
      if(!print_newline)
        printf("\033[%dA", 1);
    }
  }

  exit(EXIT_SUCCESS);
}

// Process bash options
static void process_options(int argc, char *argv[]) {

  for (;;) {
    int c = getopt(argc, argv, "hi:qmrtLR:F");

    if (c == -1) break;

    switch (c) {
      case 'm':
        clock_id = CLOCK_MONOTONIC;
        break;
      case 'r':
        clock_id = CLOCK_REALTIME;
        break;
      case 't':
        clock_id = CLOCK_TAI;
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'i':
        thread_params.interval = atoi(optarg) * 1000;
        break;
      case 'q':
        quit = 1;
        break;
      case 'L':
        print_newline = 1;
        break;
      case 'R':
        print_period = atoi(optarg);
        break;
      case 'F':
        nowait = 1;
        break;
    }
  }
}

static void add_ns(struct timespec *t, uint64_t ns) {
  t->tv_nsec += ns;

  while ((unsigned int)t->tv_nsec >= NSEC_PER_SEC) {
    t->tv_sec += 1;
    t->tv_nsec -= NSEC_PER_SEC;
  }
}
static uint64_t ts_to_uint(struct timespec t) {
  return t.tv_sec * NSEC_PER_SEC + t.tv_nsec;
}
