/*
 * Software PPS
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "gpio.h"
#include "pulse.h"

// Structs

typedef struct thread_param {
  uint64_t interval;
  unsigned int max_cycles;
  int priority;
  int affinity_cpu;
  int poll;
  int poll_margin;
  uint64_t start_ts;

} thread_param_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;
} main_param_t;

// Static functions

static void process_options(int argc, char *argv[]);
static uint64_t nb_cycles;

// Static variables

static main_param_t main_params;
static thread_param_t thread_params;
static int use_gpio;

static void help(char *argv[]) {
  printf(
      "Usage: %s [-a CPU -p PRIO -i USEC -r USEC -s NS -l N -v]\n\n"
      "       -h                 Show help\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -l N_CYCLES        RT thread cycles amount (default: infinite)\n"
      "       -s NS              Common start time reference\n"
      "       -g                 Use GPIO instead of serial\n"
      "       -P USEC            Do polling to wakeup with specified margin\n"
      "       -v                 Verbose\n"
      "\n",
      argv[0]);
}

static void poll_wakeup(struct timespec ts, int margin) {
  int ret;
  struct timespec ts_prev, current;

  ts_prev = ts;
  substract_ns(&ts_prev, margin * 1000);

  ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &ts_prev, NULL);
  if (ret) {
    fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
    exit(EXIT_FAILURE);
  }

  do {
    clock_gettime(CLOCK_REALTIME, &current);
  } while(calcdiff_ns_signed(ts, current) > 1000);
}

/*
 * Real-time thread:
 */
static void *pps_thread(void *p) {
  (void)p;
  struct timespec next;
  int ret;
  cpu_set_t mask;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu != -1) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_REALTIME, &next);

  if (thread_params.start_ts) {
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
  }

  // Packet sending loop
  for (nb_cycles = 0;; nb_cycles++) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    if (use_gpio)
      toggle_gpio();
    else
      send_pulse();

    add_ns(&next, thread_params.interval);

    if(thread_params.poll) {
      poll_wakeup(next, thread_params.poll_margin);
    }
    else {
      ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &next, NULL);
      if (ret) {
        fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
        exit(EXIT_FAILURE);
      }
    }
  }

  pthread_exit(NULL);
}

/*
 * Main thread:
 *  - Has non-real time priority
 *  - Handles the IO and creates the real time thread
 */
int main(int argc, char *argv[]) {
  pthread_t thread;
  struct sched_param param;
  pthread_attr_t attr;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.max_cycles = 0;
  thread_params.priority = 98;
  thread_params.affinity_cpu = -1;
  thread_params.start_ts = 0;
  thread_params.poll = 0;
  thread_params.poll_margin = 75;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  use_gpio = 0;

  // Lock all current and future pages from preventing of being paged to
  // swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    fprintf(stderr, "mlockall failed");
    exit(EXIT_FAILURE);
  }

  // Process bash options
  process_options(argc, argv);

  if (use_gpio)
    enable_gpio(86);
  else
    enable_pulse();

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, pps_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create pps thread");

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    if (main_params.verbose) {
      printf("Toggles: %9" PRIu64 "\n", nb_cycles);
      printf("\033[%dA", 1);
    }

    if (thread_params.max_cycles)
      if (thread_params.max_cycles == nb_cycles) break;
  }

  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  for (;;) {
    int c = getopt(argc, argv, "a:hgi:l:p:r:s:vP:");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'g':
        use_gpio = 1;
        break;
      case 'i':
        thread_params.interval = strtoull(optarg, NULL, 10) * 1000;
        break;
      case 'l':
        thread_params.max_cycles = atoi(optarg);
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'P':
        thread_params.poll = 1;
        thread_params.poll_margin = atoi(optarg);
        break;
    }
  }
}
